-- Visualization of Lazy Evaluation and Sharing
-- Chapter 2: Background

import Prelude hiding (takeWhile, map)

-- | 'takeWhile', applied to a predicate @p@ and a list @xs@, returns the
-- longest prefix (possibly empty) of @xs@ of elements that satisfy @p@:
--
-- > takeWhile (< 3) [1,2,3,4,1,2,3,4] == [1,2]
-- > takeWhile (< 9) [1,2,3] == [1,2,3]
-- > takeWhile (< 0) [1,2,3] == []
--

takeWhile               :: (a -> Bool) -> [a] -> [a]
takeWhile _ []          =  []
takeWhile p (x:xs)
            | p x       =  x : takeWhile p xs
            | otherwise =  []

map          :: (a -> b) -> [a] -> [b]
map _ []     =  []
map f (x:xs) =  f x : map f xs

squares      :: [Integer]
--squares      = map (^2) [1..10]
squares      = map (^2) [1..]

smallSquares :: [Integer]
smallSquares = takeWhile (<= 20) squares

doubleTotal = total + total
  where total = sum [1..100]

data Tree = Leaf
          | Node Tree Int Tree

t :: Tree
t = Node (Node Leaf 2 Leaf) 1 Leaf
