{-# LANGUAGE RankNTypes #-}
-- Visualization of Lazy Evaluation and Sharing
-- Chapter 4: Evaluation
-- Section 4.2: Visualizations
-- Subsection 4.2.8: Nexuses and Dynamic Programming

-- This code is from "Trouble shared is trouble halved" by R.Bird and R.Hinze

-- Note: The Empty data constructor has an argument in interpreted code to make
--       it reach the minimum size of a heap object

data Tree a = Empty
            | Node { left :: Tree a, info :: a, right :: Tree a }
  deriving (Eq, Show)

leaf           :: forall a.a -> Tree a
leaf x         = Node Empty x Empty

insert         :: forall a.Ord a => a -> Tree a -> Tree a
insert x Empty = leaf x
insert x (Node l k r)
  | x <= k    = Node (insert x l) k r
  | otherwise = Node l k (insert x r)

full           :: forall a.Integer -> a -> Tree a
full 0 x       = leaf x
full n x       = Node t x t
  where t = full (n-1) x

full'          :: forall a.Integer -> a -> Tree a
full' 0 x      = leaf x
full' n x      = Node (full' (n-1) x) x (full' (n-1) x)

memofib        :: Integer -> Tree Integer
memofib 0      = leaf 0
memofib 1      = Node (leaf 0) 1 Empty
memofib n      = node t (left t)
  where t = memofib (n-1)

node :: Tree Integer -> Tree Integer -> Tree Integer
node l r = Node l (info l + info r) r
