{-# LANGUAGE RankNTypes #-}
-- Visualization of Lazy Evaluation and Sharing
-- Chapter 4: Evaluation
-- Section 4.2: Visualizations
-- Subsection 4.2.9: Optimal Bracketing

-- This code is from "Trouble shared is trouble halved" by R.Bird and R.Hinze

-- Note: The Empty data constructor has an argument in interpreted code to make
--       it reach the minimum size of a heap object

data Tree a = Empty
            | Node { left :: Tree a, info :: a, right :: Tree a }
  deriving (Eq, Show)

data Expr a = Const a
            | Expr a :+: Expr a
  deriving Show

opt           :: [a] -> Expr a
opt [x]       = Const x
opt xs        = best [opt s1 :+: opt s2 | (s1,s2) <- uncat xs]

uncat         :: forall a.[a] -> [([a], [a])]
uncat [x1,x2] = [([x1], [x2])]
uncat (x:xs)  = ([x],xs) : map(\(l,r) -> (x:l,r)) (uncat xs)

best    :: [Expr a] -> Expr a
best xs = snd $ go xs
  where go [y]          = (size y, y)
        go (y:ys)
          | sizeY < sys = (sizeY, y)
          | otherwise   = gys
          where gys@(sys,_) = go ys
                sizeY = size y

size             :: Expr a -> Integer
size ((l :+: r)) = 1 + max (size l) (size r)
size (Const _)   = 1

leaf   :: a -> Tree (Expr a)
leaf x = Node Empty (Const x) Empty

lspine, rspine      :: forall a.Tree a -> [a]
lspine (Empty)      = []
lspine (Node l x r) = lspine l ++ [x]

rspine (Empty)      = []
rspine (Node l x r) = [x] ++ rspine r

combine             :: [(Expr a, Expr a)] -> Expr a
combine             = best . map (uncurry (:+:))

node                :: Tree (Expr a) -> Tree (Expr a) -> Tree (Expr a)
node l r            = Node l (combine (zip (lspine l) (rspine r))) r

step                :: [Tree (Expr a)] -> [Tree (Expr a)]
step [t]            = []
step (t1:t2:ts)     = node t1 t2 : step (t2:ts)

build               :: [Tree (Expr a)] -> Tree (Expr a)
build [t]           = t
build ts            = build (step ts)

bottomup            :: [a] -> Tree (Expr a)
bottomup            = build . map leaf
